@@ -202,11 +202,11 @@ class Byneuron:
             "grant_type": "client_credentials"
         }
         url = self.config.url_login
-        '''
+        """
         if url is None:
             # note this endpoint is no longer used.
             url = f'{self.api}/login'
-        '''
+        """
         r = self.url(url=url, headers=login_headers, data=data, as_json=False)
         return r
 
@@ -266,13 +266,8 @@ class Byneuron:
         :param filters:
         :return:
         """
-        if isinstance(indexsets, list):
-            if all([isinstance(i, Entity) for i in indexsets]):
-                indexsets = [i.key for i in indexsets]
-            else:
-                indexsets = indexsets
-        else:
-            indexsets = [self.indexset_active.key]
+        indexsets = [i.key if isinstance(i, Entity) else i for i in indexsets] if isinstance(indexsets, list) \
+            else [self.indexset_active.key]
         filters = [] if not isinstance(filters, list) else filters
         data = {
             "indexSets": indexsets,  # list of indexSet keys
@@ -285,13 +280,82 @@ class Byneuron:
         url = f'{self.api}/backend/rest/numberevents/_query'
         return self.url(url=url, data=data)
 
+    def _datehistogramtimeline(self, items, aggrs, from_date, to_date, interval,
+                               filters=None, gap=None, order=None, deleted=None):
+        """
+        :param items: (list of) entity of the type Item, or entity.key if in active indexset
+        :param aggrs: list of aggregations, e.g. ['avg']
+        :param from_date: date to start aggregation from
+        :param to_date: end date to end aggregation with
+        :param interval: 'bucket size' e.g. 15m
+        :param filters: optional, method to e.g. include numberValue filtering
+        :param gap: optional, specify gapstrategy
+        :param order: optional, sorting on timestamp
+        :param deleted: optional, include deleted events
+        :return:
+        """
+
+        items = items if isinstance(items, list) else [items]
+        indexsets = list({i.indexset if isinstance(i, Entity) else self.indexset_active.key for i in items})
+        timeseries = [i.key if isinstance(i, Entity) else i for i in items]
+        agg_metrics = ['avg', 'min', 'max', 'count', 'diff', 'sum', 'close', 'cardinality', 'open']
+        aggrs = [{
+            "type": "metric",
+            "name": m,
+            "stat": m,
+            "field": "numberValue"
+        } for m in aggrs if m in agg_metrics]
+        filters = filters if isinstance(filters, list) else []
+        filters.append([
+            {
+                "type": "eventitemid",
+                "values": timeseries,
+                "invertFilter": False,
+                "operator": "or"
+            },
+            {
+                "type": "range",
+                "from": int(1000 * from_date.timestamp()),
+                "to": int(1000 * to_date.timestamp()),
+                "minInclusive": True,
+                "maxInclusive": False,
+                "invertFilter": False
+            }
+        ])
+        aggregation = {
+            "type": "date",
+            "interval": interval,
+            "name": "datehistogram",
+            "from": int(1000 * from_date.timestamp()), 
+            # wierldly, this 'from' is a requirement to retrieve data with gapstrategy on
+            "to": int(1000 * to_date.timestamp()),
+            "aggregations": aggrs
+        }
+        data = {
+            "query": {
+                "indexSets": indexsets,
+                "sortField": "timestamp",
+                "order": "DESC" if order is None else order, # todo order is not respected
+                "deleted": False if deleted is None else bool(False),
+                "filters": filters,
+                "esDocumentType": "NumberEvent"
+            },
+            "gapStrategy": {} if gap is None else gap,
+            "aggregation": aggregation
+        }
+        if filters:
+            data['query']['filters'].append(filters)
+        url = f'{self.api}/backend/rest/numberevents/_datehistogramtimeline'
+        log.info('data for _datehistogramtimeline %s', data)
+        return self.url(url=url, data=data)
+
     ### endpoint handlers ###
     def get_numberevents(self, item, size):
         """ last 100 events for item in {datetime:value} format """
         items = item if isinstance(item, list) else [item]
         item_keys = list({i.key if isinstance(i, Entity) else i for i in items})
-        indexset_keys = list({i.indexset if isinstance(i, Entity) else self.indexset_active.keys for i in items})
-        log.info('get numberevents for in indexset %s items %s ', indexset_keys, item_keys)
+        indexset_keys = list({i.indexset if isinstance(i, Entity) else self.indexset_active.key for i in items})
+        log.debug('get numberevents for in indexset %s items %s ', indexset_keys, item_keys)
         f = {
             "type": "eventitemid",
             "values": item_keys,
@@ -307,6 +371,50 @@ class Byneuron:
     def get_numberevents_dict(self, item, size):
         return {e.datetime: e.value for e in self.get_numberevents(item=item, size=size)}
 
+    def get_aggregation(self, item, aggrs, interval, size, push=None):
+        """
+        retrieve x aggregated buckets of data, including current bucket
+        :param item:
+        :param aggrs: (list of) aggregation(s) ['avg']
+        :param interval: bucket size '15m'
+        :param size: number of buckets
+        :param push: boolean to set gap strategy 'push' with extended timeline
+        :return:
+        """
+        aggrs = aggrs if isinstance(aggrs, list) else [aggrs]
+        agg_intervals = {
+            '1m': datetime.timedelta(minutes=1),
+            '5m': datetime.timedelta(minutes=5),
+            '15m': datetime.timedelta(minutes=15),
+            '1h': datetime.timedelta(hours=1),
+            '1d': datetime.timedelta(hours=24),
+            #'1M': None  # todo
+        }
+
+        end = self.now
+        if interval not in agg_intervals:
+            raise ValueError('unsupported interval')
+        interval_dt = agg_intervals[interval]
+        end_rounded = end - datetime.timedelta(seconds=(end.timestamp() % interval_dt.total_seconds()))
+        start_rounded = end_rounded - (interval_dt * (size-1))
+        gap = push if push is None else {"type": "push", "extendTimeline": True}
+        r = self._datehistogramtimeline(
+            items=item,
+            aggrs=aggrs,
+            from_date=start_rounded,
+            to_date=end,
+            interval=interval,
+            gap=gap
+        )
+        if isinstance(r, list):
+            for i in r:
+                bucket = {
+                    'time': i.get('time', None),
+                    'count': i.get('count', 0)
+                }
+                bucket.update(i.get('values', {}))
+                yield bucket
+
     def query(self, query, keys):
         """
         Query a nuql expression and extract the entities for the requested variable
@@ -629,6 +737,10 @@ class Entity:
     def unit(self):
         return self._attr('unit')
 
+    @property
+    def active(self):
+        return self._attr('active')
+
     @property
     def codetype(self):  # for indexSetCode
         return self._attr('codeType')
@@ -648,6 +760,24 @@ class Entity:
             except TypeError:
                 log.exception('type error on entity data')
 
+    @property
+    def from_date(self):
+        date = self._attr('fromDate')
+        if date is not None:
+            return datetime.datetime.fromisoformat(date).replace(tzinfo=datetime.timezone.utc)
+
+    @property
+    def to_date(self):
+        date = self._attr('toDate')
+        if date is not None:
+            return datetime.datetime.fromisoformat(date).replace(tzinfo=datetime.timezone.utc)
+
+    @property
+    def meta_indexset_codes(self):
+        meta = self._attr('metaIndexSetCodes')
+        if meta:
+            return {i.get('key'): i.get('value') for i in meta}
+
     @property
     def meta_keywords(self):
         meta = self._attr('metaKeywords')
@@ -694,6 +824,14 @@ class Entity:
     def items(self):
         return [i for i in self._links('hasFeatureItem')]
 
+    @property
+    def source(self):
+        return self._link('hasSource')
+
+    @property
+    def target(self):
+        return self._link('hasTarget')
+
     def create_event(self, value, dt):
         event = Numberevent()
         event.create(value=value, dt=dt, item=self)
@@ -822,7 +960,7 @@ class Entity:
         except (Exception,):
             log.exception('set_attribute')
 
-    '''
+    """
     # idea to allow event creation directly from entity
     def set_event(self, val, dt=None):
         #test if byneuron is defined
@@ -832,7 +970,7 @@ class Entity:
         event = Numberevent()
         event.create(value=val, dt=dt, item=self)
         self.bn.graphmodel(numberEvents=[event])
-    '''
+    """
 
 
 class Numberevent:
